/***********************************************************************/
/*                                                                     */
/*                                                                     */
/* Program                                                             */
/* -------                                                             */
/*                                                                     */
/* dcpend1.c for Velleman K8061.                                       */
/*                                                                     */
/*                                                                     */
/* Purpose                                                             */
/* -------                                                             */
/*                                                                     */
/* This is a simple PID-regulator program to control inverted          */
/* pendulums and other similar things from a PC. The Velleman K8061    */
/* USB I/O-board is used as an interface. A potentiometer (10 kOhm)    */
/* for the pendulum direction and an auxiliary board whith some        */
/* operational amplifiers are also needed. The Devantech MD03          */
/* (max 50 V, 20 A) is used as motor controller. Some auxiliary        */
/* power supplies (5, 15 V 200 mA and 18 V 3 A) are also needed.       */
/*                                                                     */
/*                                                                     */
/* Version 1.03                                                        */
/* ------------                                                        */
/*                                                                     */
/* First version: 2009-08-26                                           */
/* Latest update: 2009-08-29                                           */
/*                                                                     */
/*                                                                     */
/* Author of the program                                               */
/* ---------------------                                               */
/*                                                                     */
/* Stefan Spaennare, Lund, Sweden                                      */
/*                                                                     */
/* E-mail: stefan@spaennare.se                                         */
/*                                                                     */
/* Home page:     http://www.spaennare.se/index.html                   */
/* Project page:  http://www.spaennare.se/dcpend1.html                 */
/* Software page: http://www.spaennare.se/dcpend1.html                 */
/*                                                                     */
/*                                                                     */
/* Documentation and hints                                             */
/* -----------------------                                             */
/*                                                                     */
/* More documentation is found on the authors project page and in      */
/* the readme file that comes with this package.                       */
/*                                                                     */
/* It is recommended to have the K8061 board USB 2.0 connected for     */
/* fastest performance. Then times down to 1.0 ms per I/O-operation    */
/* can be achieved. A PC computer >= 1.0 GHz i recommended.            */
/*                                                                     */
/*                                                                     */
/* Input parameters                                                    */
/* ----------------                                                    */
/*                                                                     */
/* Usage: dcpend1 h0-msr h-iti pgainr igainr dgainr                    */
/*                                                                     */
/* Parameters ending with "i" are integers and "r" are real.           */
/*                                                                     */
/* h0-msr: Iteration delay in ms. About 5.0 ms is recommended here.    */
/* h-iti:  Integer iteration delay loop. Should be adjusted to h0-ms   */
/*         as exact as possible. A value between 60000 and 300000 is   */
/*         usually suitable.                                           */
/*                                                                     */
/* pgainr: Proportional gain. Start with 6.0.                          */
/* igainr: Integrating gain.  Start with 6.0.                          */
/* dgainr: Derivative gain.   Start with 1.2.                          */
/*                                                                     */
/* The PID-parameters can then be adjusted for optimal performance.    */
/* These parameters also depend on the pendulum length and mass.       */
/*                                                                     */
/* The author found that it is better to have a fix preset "h0-msr"    */
/* that is correct on average (over 1 second or more) than to have     */
/* it adaptivley adjusted to "h-real" for each iteration.              */ 
/*                                                                     */
/*                                                                     */
/* Reference value                                                     */
/* ---------------                                                     */
/*                                                                     */
/* Adjust the reference value (ref0) for the pendulum direction        */
/* potentiometer from the computer keyboard.                           */
/*                                                                     */
/* s =>   +1 ; a =>   -1 to ref                                        */
/* f =>  +10 ; d =>  -10 to ref                                        */
/* h =>  +25 ; g =>  -25 to ref                                        */
/* x =>  +50 ; z =>  -50 to ref                                        */
/* v => +100 ; c => -100 to ref                                        */
/* n => +200 ; b => -200 to ref                                        */
/* space => back to ref0                                              */
/* q => quit program                                                  */
/*                                                                     */
/*                                                                     */
/* Config file                                                         */
/* -----------                                                         */
/*                                                                     */
/* Some auxiliary parameters can be set in the "dcpend1.cfg" config    */
/* file. See this file for more information.                           */
/*                                                                     */
/*                                                                     */
/* Input and output                                                    */
/* ----------------                                                    */
/*                                                                     */
/* AD-channel 1 (0-1023, 0-5 V) on the Velleman K8061 I/O-board is     */
/* used for potentiometer feedback (via an operational amplifier       */
/* follower). DA-channel 1 (0-255, 0-5 V) on K8061 is used for         */
/* speed output to SDA input on the Devantech MD03 motor controller.   */
/* Digital output 1 (0 or +5 V, open collector) on K8061 is used as    */
/* direction input to SCL on the MD03 motor contoller via a 1 kOhm     */ 
/* pull up resitor to +5 V.                                            */
/*                                                                     */
/* If you wish to have more inputs from potentiometers or other        */
/* actuators you can use AD-channel 2 - 7 on the Velleman K8061        */
/* I/O-board. Then this C-program must also be somewhat modified       */
/* and recompiled.                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

/********************************************************************************/
/*                                                                              */
/* Notice                                                                       */
/* ======                                                                       */
/*                                                                              */
/* I make no warranties that this program is (1) free of errors, (2) consistent */
/* with any standard merchantability, or (3) meeting the requirements of a      */
/* particular application. This software shall not, partly or as a whole,       */
/* participate in a process, whose outcome can result in injury to a person or  */
/* loss of property. It is solely designed for analytical work. Permission to   */
/* use, copy and distribute is hereby granted without fee, providing that the   */
/* header above including this notice appears in all copies.                    */
/*                                                                              */
/*                                                            Stefan Spaennare  */
/*                                                                              */
/********************************************************************************/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <time.h>
#include <windows.h>
#include "K8061.h"


double timetic=(double)(CLOCKS_PER_SEC);

char cfg_name[21]="dcpend1.cfg";


double harm_delay(long delay0)
{
   long j;
   double sum;


      sum=0.0;

      for (j=1; j<=delay0; j++) {
         sum=sum+1.0/(double)(j);
      } /* for j */

      return(sum);

} /* harm_delay */


long ref_adjust(char cc)
{
   long temp;

      temp=0;

      if (cc == 's') {
         temp=1;
      } /* if */

      if (cc == 'a') {
         temp=-1;
      } /* if */

      if (cc == 'f') {
         temp=10;
      } /* if */

      if (cc == 'd') {
         temp=-10;
      } /* if */

      if (cc == 'h') {
         temp=25;
      } /* if */

      if (cc == 'g') {
         temp=-25;
      } /* if */

      if (cc == 'x') {
         temp=50;
      } /* if */

      if (cc == 'z') {
         temp=-50;
      } /* if */

      if (cc == 'v') {
         temp=100;
      } /* if */

      if (cc == 'c') {
         temp=-100;
      } /* if */

      if (cc == 'n') {
         temp=200;
      } /* if */

      if (cc == 'b') {
         temp=-200;
      } /* if */

      return(temp);

} /* ref_adjust */


int main(int argc,char *argv[])
{

   FILE *cfgfile;

   long i,j,k,kk,flag,print_step,clock_step,aver_step;
   long data,data_in1,data_ref,ref0,data0,sgn0;
   long card_adr,analog_in_channel1,digital_out_channel1;
   long off1a,off1b,h_it,analog_out_channel1,moff0;

   double c1,c2,c3,c4,h_time,h,h0,h_inv;
   double igain,pgain,dgain,error0,error_old;
   double pterm,dterm,iterm,u,dummy,u_mean,in_mean;
   double aver_inv,u_1,in_1,cs_tic_inv; 

   char cc;

   clock_t time1,time2,timediff1;

   if (argc != 6) {
      printf("Usage: %s h0-msr h-iti pgainr igainr dgainr\n",argv[0]);
      printf("\n");
      printf("s =>   +1 ; a =>   -1 to ref\n");
      printf("f =>  +10 ; d =>  -10 to ref\n");
      printf("h =>  +25 ; g =>  -25 to ref\n");
      printf("x =>  +50 ; z =>  -50 to ref\n");
      printf("v => +100 ; c => -100 to ref\n");
      printf("n => +200 ; b => -200 to ref\n");
      printf("space => back to ref0\n");
      printf("q => quit program\n");
      printf("\n");
      exit(0);
   } /* if */

   h0=atof(argv[1]);
   h_it=atoi(argv[2]);
   pgain=atof(argv[3]);
   igain=atof(argv[4]);
   dgain=atof(argv[5]);

   if ((cfgfile = fopen(cfg_name,"r")) == NULL) {
      printf("File %s not found.\n",cfg_name);
      exit(0);
   } /* if */

   fscanf(cfgfile,"%d",&off1a);
   fscanf(cfgfile,"%d",&off1b);
   fscanf(cfgfile,"%d",&ref0);
   fscanf(cfgfile,"%d",&sgn0);
   fscanf(cfgfile,"%d",&moff0);
   fscanf(cfgfile,"%d",&aver_step);
   fscanf(cfgfile,"%d",&print_step);
   fscanf(cfgfile,"%d",&clock_step);

   if (sgn0 >=0) {
      sgn0=1;
   }
   else {
       sgn0=-1;
   } /* if */

   if (aver_step < 1) {
      aver_step=1;
   } /* if */

   if (print_step < 1) {
      print_step=1;
   } /* if */

   if (clock_step < 1) {
      clock_step=1;
   } /* if */

   printf("\n");

   card_adr=0;
   analog_in_channel1=1;

   OpenDevice();

   digital_out_channel1=1;
   SetDigitalChannel(card_adr,digital_out_channel1);

   data=0;
   analog_out_channel1=1;
   OutputAnalogChannel(card_adr,analog_out_channel1,data);

   printf("\n");

   c1=1.0/10.23;
   c2=10.23;
   c3=1.0/2.55;
   c4=2.55;

   data_ref=ref0;

   analog_in_channel1=1;
   data_in1=ReadAnalogChannel(card_adr,analog_in_channel1);

   error_old=c1*(double)(data_ref)-c1*(double)(data_in1);

   dummy=0.0;

   dummy=dummy+harm_delay(h_it);  

   aver_inv=1.0/(double)(aver_step);

   cs_tic_inv=1.0/((double)(clock_step)*timetic);

   iterm=0.0;

   h0=0.001*h0;
   h=h0;
   h_inv=1.0/h;

   h_time=0.0;
   u_1=0.0;
   in_1=0.0;

   i=1;
   j=1;
   k=1;
   kk=1;

   flag=0;

   cc='#';

   time1=clock();


   while (flag==0) {

      if (kbhit()) {
         cc=(getch());
      } /* if */

      if (cc != '#') {

         if (cc == 'q') {
            flag=1;
         } 
         else {
            if (cc == ' ') {
               data_ref=ref0;
            }
            else {
               data_ref=data_ref+ref_adjust(cc);
            } /* if */
         } /* if */

         cc='#';

      } /* if */

      if (data_ref < 0) {
         data_ref=0;
      } /* if */

      if (data_ref > 1023) {
         data_ref=1023;
      } /* if */

      analog_in_channel1=1;
      data_in1=ReadAnalogChannel(card_adr,analog_in_channel1);

      error0=c1*(double)(data_ref)-c1*(double)(data_in1);

      pterm=pgain*error0;

      iterm=iterm+igain*error0*h;

      dterm=dgain*(error0-error_old)*h_inv;

      error_old=error0;

      u=pterm+iterm+dterm;

      if (u > 100.0) {
         u=100.0;
      } /* if */

      if (u < -100.0) {
         u=-100.0;
      } /* if */

      u_1=u_1+u;

      in_1=in_1+(double)(data_in1);

      data0=(long)((double)(sgn0)*c4*u+0.5);

      if (data0 > 255) {
         data0=255;
      } /* if */

      if (data0 < -255) {
         data0=-255;
      } /* if */

      data=abs(data0)+moff0;

      if (data > 255) {
         data=255;
      } /* if */

      if ((data_in1 < off1a) || (data_in1 > off1b)) {
         iterm=0.0;
         u=0.0;
         data=0;
      } /* if */

      if (data0 >= 0) {
         digital_out_channel1=1;
         SetDigitalChannel(card_adr,digital_out_channel1);
      }
      else {
         digital_out_channel1=1;
         ClearDigitalChannel(card_adr,digital_out_channel1);
      } /* if */

      analog_out_channel1=1;
      OutputAnalogChannel(card_adr,analog_out_channel1,data);

      if (kk == aver_step) {
         in_mean=aver_inv*in_1;
         u_mean=aver_inv*u_1;
         in_1=0.0;
         u_1=0.0;
         kk=0;
      } /* if */

      if (k == clock_step) {
         time2=clock();
         timediff1=time2-time1;
         time1=time2;
         h_time=1000.0*(double)(timediff1)*cs_tic_inv;
         k=0;
      } /* if */

      if (j == print_step) {
         printf("| it: %8d | h-real: %6.2f ms | ref: %6d | in: %6d | out: %8.2f |\n",
         i,h_time,data_ref,(int)(in_mean+0.5),u_mean);
         j=0;
      } /* if */

      dummy=dummy+harm_delay(h_it);  

      i++;
      j++;
      k++;
      kk++;

   } /* while */


   printf("\n");

   digital_out_channel1=1;
   SetDigitalChannel(card_adr,digital_out_channel1);

   data=0;
   analog_out_channel1=1;
   OutputAnalogChannel(card_adr,analog_out_channel1,data);

   CloseDevices;

   fclose(cfgfile);

   printf("\n");
   printf("%8.3f\n",dummy);
   printf("\n");


} /* End */
